//----------------------------------------------------------------------------
//
// File:        decode.cpp
// Date:        10-Aug-2000
// Programmer:  Marc Rousseau
//
// Description:
//
// Copyright (c) 1994-2001 Marc Rousseau, All Rights Reserved.
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307, USA.
//
// Revision History:
//
//----------------------------------------------------------------------------

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "common.hpp"

#define CHUNK_FMT	0x20746D66	// "fmt "
#define CHUNK_DATA	0x61746164	// "data"

struct sRIFF_Block {
    char  Tag [ 4 ];
    ULONG Length;
};

struct sFormatChunk : sRIFF_Block {
    USHORT formatTag;		// 1 = PCM
    USHORT nChannels;
    ULONG  nSamplesPerSec;
    ULONG  nAvgBytesPerSec;
    USHORT nBlockAlign;
};

struct sFileData {
    UCHAR  *pData;
    ULONG   DataLength;
    double  period;
};

sFormatChunk *fmt;

bool CheckHeader ( UCHAR *pFileData, int fileSize, UCHAR **pData, int *pDataSize, sFormatChunk **pFormat )
{
    sRIFF_Block *pRIFF = ( sRIFF_Block * ) pFileData;

    *pData     = NULL;
    *pDataSize = 0;
    *pFormat   = NULL;

    if ( strncmp ( pRIFF->Tag, "RIFF", 4 ) != 0 ) {
        return false;
    }

    do {

        char *ptr = ( char * ) ( pRIFF + 1 );

        sFormatChunk *fmt  = NULL;
        sRIFF_Block  *data = NULL;

        if ( strncmp ( ptr, "WAVE", 4 ) == 0 ) {
            ptr += 4;
            do {
                sRIFF_Block *pChunk = ( sRIFF_Block * ) ptr;
                switch ( * ( ULONG * ) pChunk->Tag ) {
                    case CHUNK_FMT :
                        fmt = ( sFormatChunk * ) pChunk;
                        break;
                    case CHUNK_DATA :
                        data = ( sRIFF_Block * ) pChunk;
                        break;
                    default :
                        break;
                }
                if (( fmt != NULL ) && ( data != NULL )) {
                    *pData     = ( UCHAR * ) ( data + 1 );
                    *pDataSize = data->Length;
                    *pFormat   = fmt;
                    return true;
                }
                ptr = ( char * ) ( pChunk + 1 ) + (( pChunk->Length + 1 ) & 0xFFFFFFFE );
            } while ( ptr < (( char * ) ( pRIFF + 1 ) + pRIFF->Length ));
        }

        pRIFF = ( sRIFF_Block * ) (( UCHAR * ) ( pRIFF + 1 ) + (( pRIFF->Length + 1 ) & 0xFFFFFFFE ));

    } while (( UCHAR * ) pRIFF < pFileData + fileSize );

    return false;
}

int (*GetSample) ();

UCHAR *tapeBuffer;
UCHAR *endBuffer;

int GetSampleMono8 ()
{
    return * ( UCHAR * ) tapeBuffer++ - 0x80;
}

int GetSampleMono16 ()
{
    return * (( short *& ) tapeBuffer )++;
}

int GetSampleStereo8 ()
{
    int value = * ( UCHAR * ) tapeBuffer - 0x80;
    tapeBuffer += fmt->nBlockAlign;
    return value;
}

int GetSampleStereo16 ()
{
    int value = * ( short * ) tapeBuffer;
    tapeBuffer += fmt->nBlockAlign;
    return value;
}

double period, blip, min, max, threshold;
int    cutOff = 5;

#define SYNC_PATTERN_MIN    25
#define SYNC_PATTERN_MATCH  5000

int FindNextCrossing ()
{
    static int lastValue = 0;
    int value = 0;

    UCHAR *pStart = tapeBuffer;

    if ( lastValue < 0 ) {
        while ( tapeBuffer < endBuffer ) {
            value = GetSample ();
            if ( value >= cutOff ) break;
        }
    } else {
        while ( tapeBuffer < endBuffer ) {
            value = GetSample ();
            if ( value <= -cutOff ) break;
        }
    }

    lastValue = value;

    return ( tapeBuffer >= endBuffer ) ? -1 : ( tapeBuffer - pStart );
}

double FindPeriod ( UCHAR *buffer, int length )
{
    int xx = cutOff;
    cutOff = 10;

    tapeBuffer = buffer;
    endBuffer  = tapeBuffer + length;

    // Get to the first zero crossing
    FindNextCrossing ();

    int avg = FindNextCrossing ();

    // Look for the beginning of consecutive 0 bits
    for ( int x = 1; x < SYNC_PATTERN_MIN; x++ ) {
        int dist = FindNextCrossing ();
        if ( dist == -1 ) return 0.0;
        if (( dist > ( avg * 2 / x )) || ( dist < ( avg / 2 / x ))) {
            x = 1;
            avg = 0;
        }
        avg += dist;
    }

    // There should be 6144 total - we've used a few already and may have missed more
    int count = 0;
    for ( int x = 0; x < SYNC_PATTERN_MATCH; x++ ) {
        count += FindNextCrossing ();
    }

    double period = count / ( double ) SYNC_PATTERN_MATCH;

    cutOff = xx;

    return period;
}

int ReadBit ()
{
    // Residual distance from last bit
    static double d0 = 0.0;

    double distance = d0;

    int d [16], i = 0;

    d [0] = ( int ) d0;

//printf ( "d0=%10.5f", distance );

    while ( distance < min ) {
        int dist = FindNextCrossing ();
        if ( dist < 0 ) return -1;
        distance += d [++i] = dist;
//printf ( " %d", d [i] );
    }

    d0 = ( distance > max ) ? distance - max : 0;

    for ( int x = 0; x <= i; x++ ) {
        if ( d [x] >= max / 2.0 ) {
//            d0 = distance - period;
            return 0;
        }
    }

//    d0 = distance - period;

    return 1;
}

int FillBitBuffer ( UCHAR *buffer, int length, double per, UCHAR *bitBuffer )
{
    tapeBuffer = buffer;
    endBuffer  = tapeBuffer + length;

    period = per;
    blip   = period * threshold;
    min    = period * ( 1.0 - threshold );
    max    = period * ( 1.0 + threshold );

printf ( "blip   = %d\n", ( int ) blip );
printf ( "cutOff = %d\n", ( int ) cutOff );
printf ( "period = %f\n", period );
printf ( "min    = %f\n", min );
printf ( "max    = %f\n", max );

    int count = 0;

    while ( tapeBuffer < endBuffer ) {

        int bit = ReadBit ();

        bitBuffer [count++] = bit;
/*
        switch ( bit ) {
            case 0   : printf ( "0" );  break;
            case 1   : printf ( "1" );  break;
            case -1  : printf ( "@" );  break;
            case -2  : printf ( "#" );  break;
            case -3  : printf ( "*" );  break;
            case -4  : printf ( "?" );  break;
            case -5  : printf ( "+" );  break;
            default  : printf ( "-" );  break;
        }

        if ( count % 8  == 0 ) printf ( " " );
        if ( count % 64 == 0 ) printf ( "\n" );
*/
    }

printf ( "\n" );
printf ( "%d bits read\n", count );

    return count;
}

//
// Format of recorded data: Miller Code - a simple Binary Line Code
//
//        Each bit cell contains either a '0' or '1'.  For a '0' value, the
//   recorded polarity remains constant for the duration of the bit cell.  A
//   '1' value is indicated by a transition at the mid-point of the bit cell.
//   The polarity is reversed at each cell boundary.
//
// |  Header   |Byte |  Preamble |       Data       |ck|
// |  >0300 |  |Count| 8 Bytes|  |     64 Bytes     |  |
//  00 -- 00 FF xx xx 00 -- 00 FF **** User Data *** yy 
//                   [the data section is repeted twice]
//
//      xx - # of 64 byte blocks to follow
//      yy - 8-bit checksum
//

UCHAR *tapeData;

UCHAR *tapeSection [1+2+256*2*2+1];
int    nextSection;

inline int GetBit () { return *tapeData++; }

bool FindStart ()
{
restart:

    tapeSection [nextSection] = tapeData;

    // Read 0x0300 copies of 0x00
    for ( EVER ) {
        int bit = GetBit ();
        if ( bit < 0 ) return false;
        if ( bit == 1 ) break;
    }

    // Look for 0xFF
    for ( int i = 1; i < 8; i++ ) {
        int bit = GetBit ();
        if ( bit != 1 ) goto restart;
    }

    return true;
}

int GetByte ()
{
    int byte = 0;
    for ( int i = 0; i < 8; i++ ) {
        int bit = GetBit ();
        if ( bit < 0 ) return -1;
//printf ( bit ? "1" : "0" );
        byte = ( byte << 1 ) | bit;
    }
//printf ( " - %02X\n", byte );
//printf ( " %02X", byte );

    return byte;
}

bool FindPreamble ()
{
    for ( EVER ) {

        tapeSection [nextSection] = tapeData;

        int bit = 0, count = 0;

        // Look for 8 0x00's
        do {
            bit = GetBit ();
            if ( bit < 0 ) return false;
            count++;
        } while ( bit == 0 );

        if ( count > 56 ) break;
    }

    // Look for 0xFF
    for ( int i = 1; i < 8; i++ ) {
        int bit = GetBit ();
        if ( bit != 1 ) return false;
    }

    return true;
}

void dumpSection ( const char *name, UCHAR *ptr, int size )
{
    printf ( "%s: %4d - ", name, size );
    while ( size-- > 0 ) printf (( *ptr++ == 1 ) ? "1" : "0" );
    printf ( "\n" );
}

bool FindChunk ()
{
    // Find the preamble boundaries
    if ( FindPreamble () == false ) return false;
    nextSection++;

    // Record the start of data
    tapeSection [nextSection] = tapeData;
    nextSection++;

    tapeData += ( 64 + 1 ) * 8;

    return true;
}

bool ReadChunk ( int index, UCHAR *data )
{
    tapeData = tapeSection [index];

    int sum = 0;
    for ( int j = 0; j < 64; j++ ) {
        sum += data [j] = GetByte ();
    }
    int actual = GetByte ();

//    return (( sum & 0xFF ) == actual ) ? true : false;
    return (( sum & 0xFF ) == actual ) ? false : false;
}

#define MIN_RUN_SIZE    1

struct sRunInfo {
    int start;
    int length;
    int shift;
};

struct sRunQueue {
    int            count;
    sRunInfo       info [1024];
};

sRunQueue runQueue [2];

void AddRun ( sRunQueue *queue, int start, int length, int shift )
{
    if ( start < 0 ) {
        if ( start + length <= 0 ) return;
        length = start + length;
        start  = 0;
    }

    for ( int i = 0; i < queue->count; i++ ) {
        sRunInfo *info = &queue->info [i];
        if (( start == info->start ) && ( length == info->length )) {
            if ( abs ( shift ) < abs ( info->shift )) info->shift = shift;
            return;
        }
    }

if ( queue->count > ( int ) SIZE ( queue->info )) printf ( "QUEUE OVERFLOW!\n" );

    sRunInfo *info = &queue->info [queue->count++];
    info->start  = start;
    info->length = length;
    info->shift  = shift;
}

typedef int (*QSORT_FUNC) ( const void *, const void * );

int sortFunction1 ( const sRunInfo *p1, const sRunInfo *p2 )
{
    int delta;
    
    delta = p2->length - p1->length;
    if ( delta != 0 ) return delta;

    delta = p1->start - p2->start;
    if ( delta != 0 ) return delta;

    return abs ( p1->shift ) - abs ( p2->shift );
}

int sortFunction2 ( const sRunInfo *p1, const sRunInfo *p2 )
{
    int delta;
    
    delta = p1->start - p2->start;
    if ( delta != 0 ) return delta;

    return abs ( p1->shift ) - abs ( p2->shift );
}

void TrimRunQueue ( sRunQueue *queue )
{
    for ( int i = 0; i < queue->count; i++ ) {
        sRunInfo *srcInfo = &queue->info [i];
        for ( int j = i + 1; j < queue->count; j++ ) {
            sRunInfo *dstInfo = &queue->info [j];
            if (( srcInfo->start >= dstInfo->start ) && ( srcInfo->start <= dstInfo->start + dstInfo->length )) {
                dstInfo->length = srcInfo->start - dstInfo->start;
            }
            if (( dstInfo->start >= srcInfo->start ) && ( dstInfo->start <= srcInfo->start + srcInfo->length )) {
                dstInfo->length = dstInfo->start + dstInfo->length - ( srcInfo->start + srcInfo->length );
                dstInfo->start  = srcInfo->start + srcInfo->length;
            }
            if ( dstInfo->length <= 0 ) {
                memcpy ( dstInfo, dstInfo + 1, sizeof ( sRunInfo ) * ( queue->count - j - 1 ));
                queue->count--;
                j--;
            }
        }
    }
}

// 
// We have a (very poor) rate 1/2 code (actually a rate 32/65), so we might be able to recover from minor errors
//
bool RecoverChunk ( int index1, int index2, UCHAR *data )
{
    UCHAR *src1 = tapeSection [index1];
    int length1 = tapeSection [index1+1] - tapeSection [index1];
    UCHAR *src2 = tapeSection [index2];
    int length2 = tapeSection [index2+1] - tapeSection [index2];

    // Look for the simple case - no dropped/added bits and less than a bytes worth of damage
    if (( length1 == length2 ) && ( length1 == 8 * ( 64 + 1 ))) {

        int errors = length1;
        for ( int i = 0; i < length1; i++ ) {
            if ( src1 [i] == src2 [i] ) errors--;
        }

        if ( errors <= 8 ) {
            // TBD - fixem - make sure there is at most 1 bit in error at each bit position
            printf ( "-- possibly correctable errors detected --\n" );
//            return false;
        }
    }

printf ( "Comparing chunks (%d/%d)\n", length1, length2 );
for ( int i = 0; i < length1; i++ ) printf ( "%c", '0' + src1 [i] ); printf ( "\n" );
for ( int i = 0; i < length2; i++ ) printf ( "%c", '0' + src2 [i] ); printf ( "\n" );

    runQueue [0].count = 0;
    runQueue [1].count = 0;

    // We need to find out how much of the two copies match each other
    // By shifting them around, we can detect/compensate for dropped bits
    for ( int offset = -8; offset < 8; offset++ ) {

        UCHAR *ptr1 = ( offset > 0 ) ? src1 + offset : src1;
        UCHAR *ptr2 = ( offset < 0 ) ? src2 - offset : src2;

        int max = ( offset < 0 ) ? 
                  ( length2 + offset < length1 ) ? length2 + offset : length1 :
                  ( length1 - offset < length2 ) ? length1 - offset : length2;

        int matches = 0, run = 0;
        for ( int i = 0; i < max; i++ ) {
            if ( *ptr1++ == *ptr2++ ) {
                matches++;
                run++;
            } else {
                int start  = i - run;
                if ( run > MIN_RUN_SIZE ) {
                    AddRun ( &runQueue [0], ( offset > 0 ) ? start - offset : start, run, offset );
                    AddRun ( &runQueue [1], ( offset < 0 ) ? start - offset : start, run, offset );
                }
                run = 0;
            }
        }
        int start = max - run;
        AddRun ( &runQueue [0], ( offset > 0 ) ? start - offset : start, run, offset );
        AddRun ( &runQueue [1], ( offset < 0 ) ? start - offset : start, run, offset );

        // Look for the simple case - no dropped/added bits and less than a bytes worth of damage
        if (( length1 == length2 ) && ( length1 == 8 * ( 64 + 1 )) && ( length1 - matches <= 8 )) {
        }
    }

    qsort ( runQueue[0].info, runQueue[0].count, sizeof ( sRunInfo ), ( QSORT_FUNC ) sortFunction1 );
    TrimRunQueue ( &runQueue [0] );
    qsort ( runQueue[0].info, runQueue[0].count, sizeof ( sRunInfo ), ( QSORT_FUNC ) sortFunction2 );

    qsort ( runQueue[1].info, runQueue[1].count, sizeof ( sRunInfo ), ( QSORT_FUNC ) sortFunction1 );
    TrimRunQueue ( &runQueue [1] );
    qsort ( runQueue[1].info, runQueue[1].count, sizeof ( sRunInfo ), ( QSORT_FUNC ) sortFunction2 );

for ( int i = 0; i < 2; i++ ) {
    printf ( "runQueue[%d]: %d entries\n", i, runQueue [i].count );
    int bits = 0;
    for ( int x = 0; x < runQueue[i].count; x++ ) {
        printf ( " %3d - %3d, %3d\n", runQueue [i].info [x].start, runQueue [i].info [x].length, runQueue [i].info [x].shift );
        bits += runQueue [i].info[x].length;
    }
    printf ( "chunk[%d]: %3d good bits (short by %d)\n", i, bits, 65 * 8 - bits );
}

    return false;
}

/*
bool ReadChunk ( int index, UCHAR *data )
{
    char checksum1 = -1, checksum2 = -2, shadow [64];
    bool good = false;

    for ( int i = 0; i < 2; i++ ) {
        if ( FindPreamble () == false ) return good;
        nextSection++;
        tapeSection [nextSection] = tapeData;
        nextSection++;
        int sum = 0;
        for ( int j = 0; j < 64; j++ ) {
            sum += data [j] = GetByte ();
        }
        int actual = GetByte ();
        if (( sum & 0xFF ) == actual ) good = true;
        if ( i == 0 ) {
            memcpy ( shadow, data, 64 );
            checksum1 = actual;
        } else {
            checksum2 = actual;
        }
    }

    printf (  " preamble[0]: " );  dumpSection ( tapeSection [nextSection-4], tapeSection [nextSection-3] - tapeSection [nextSection-4] );
    printf ( "      data[0]: " );  dumpSection ( tapeSection [nextSection-3], tapeSection [nextSection-2] - tapeSection [nextSection-3] );
    printf ( "  preamble[1]: " );  dumpSection ( tapeSection [nextSection-2], tapeSection [nextSection-1] - tapeSection [nextSection-2] );
    printf ( "      data[1]: " );  dumpSection ( tapeSection [nextSection-1], tapeData - tapeSection [nextSection-1] );


    // See if we can recover a single byte error
    if (( good == false ) && ( checksum1 == checksum2 )) {
        int index = 0, errors = 0;
        int sum = 0;
        for ( int j = 0; j < 64; j++ ) {
            sum += data [j];
            if ( data [j] != shadow [j] ) {
                errors++;
                index = j;
            }
        }
        if ( errors == 1 ) {
            data [index] = checksum1 - ( sum - data [index]);
            good = true;
        }
    }

    return good;
}
*/

bool DumpBuffer ( UCHAR *buffer, int size, const char *filename )
{
    // Set up the globals and EOF condition
    tapeData = buffer;
    tapeData [size] = ( UCHAR ) -1;

    // Start at the beginnng...
    nextSection = 0;

    if ( FindStart () == false ) {
        fprintf ( stderr, "Unable to locate start of data\n" );
        return false;
    }

    nextSection++;
    tapeSection [nextSection] = tapeData;
    nextSection++;

    // Read count
    int count1 = GetByte ();
    int count2 = GetByte ();

    if (( count1 < 0 ) || ( count2 < 0 )) {
        return false;
    }

    if ( count1 != count2 ) {
        fprintf ( stderr, "Counts do not match!\n" );
        return false;
    }

    UCHAR *data = new UCHAR [ count1 * 64 ];

    printf ( "Reading %d chunks: ", count1 );

    bool good = true;
    for ( int i = 0; i < count1; i++ ) {
         FindChunk ();
         FindChunk ();
    }

    tapeSection [nextSection] = tapeData;

/*
for ( int i = 0; i < count1; i++ ) {
    int index = 2 + i * 4;
    dumpSection ( "preamble[0]", tapeSection [index], tapeSection [index+1] - tapeSection [index] ); index++;
    dumpSection ( "    data[0]", tapeSection [index], tapeSection [index+1] - tapeSection [index] ); index++;
    dumpSection ( "preamble[1]", tapeSection [index], tapeSection [index+1] - tapeSection [index] ); index++;
    dumpSection ( "    data[1]", tapeSection [index], tapeSection [index+1] - tapeSection [index] ); index++;
}
*/

    for ( int i = 0; i < count1; i++ ) {
        int index1 = 2 + 1 + i * 4;
        if ( ReadChunk ( index1, data + i * 64 ) == false ) {
            int index2 = 2 + 1 + i * 4 + 2;
            if ( ReadChunk ( index2, data + i * 64 ) == false ) {
                if ( RecoverChunk ( index1, index2, data + i * 64 ) == false ) {
                    good = false;
                    printf ( "x" );
                    continue;
                }
                printf ( "*" );
                continue;
            }
        }
        printf ( "." );
    }

/*
    if ( good == true ) {
        FILE *file = fopen ( filename, "wb" );
        fwrite ( data, count1, 64, file );
        fclose ( file );
    }
*/
    printf ( "\n" );

    delete [] data;

    return good;
}

bool debug = false;

int LocateFiles ( UCHAR *pData, ULONG dataSize, sFileData *fileData )
{
    enum {
        STAGE_UNKNOWN,
        STAGE_PREAMBLE,
        STAGE_DATA,
    } stage = STAGE_UNKNOWN;

    int count = 0;

    int periodCount = 0;
    double periodSum = 0.0;

    tapeBuffer = pData;
    endBuffer  = tapeBuffer + dataSize;

    fileData [ count ].pData = tapeBuffer;

    const int PREAMBLE_SIZE = 0x300 * 8;
    const int PREAMBLE_THRESHOLD = 2 * PREAMBLE_SIZE / 3;

    int preSize = 0, bad = 0;

    do {

        int x;
        int sum     = 0;
        int squares = 0;

        for ( x = 0; x < 50; x++ ) {
            int dist = FindNextCrossing ();
            if ( dist == -1 ) goto done;
            sum     += dist;
            squares += dist * dist;
        }
        double avg = ( double ) sum / ( double ) x;
        double var = ( double ) squares / x - avg * avg;

        double threshold = avg * avg;

if ( debug ) printf ( "var = %10.5f  avg = %10.5f - %10.5f (%d)\n", var, avg, var / ( avg * avg ), preSize );

        if ( var < 0.1 * threshold ) {
            if (( var <= 1.0 ) && ( preSize > 5 )) {
                periodSum += avg;
                periodCount++;
            }
            if ( stage == STAGE_UNKNOWN ) {
if ( debug ) printf ( "Start of preamble detected\n" );
                stage = STAGE_PREAMBLE;
                preSize = 0;
            }
            bad = 0;
            preSize++;
        } else if (( stage == STAGE_PREAMBLE ) && ( var < 0.6 * threshold )) {
if ( debug ) printf ( "-- noise in preamble detected\n" );
            bad++;
            if (( bad >= 1 ) && ( preSize > PREAMBLE_THRESHOLD / x )) {
if ( debug ) printf ( "Switching to data\n" );
                stage = STAGE_DATA;
                preSize = 0;
            } else if ( bad < 15 ) {
//                preSize++;
            } else {
//if ( debug ) printf ( "Too much noise - preamble reset\n" );
//                preSize = 0;
            }
//        } else if ( var < 5 * avg ) {
        } else if ( var < 1.0 * threshold ) {
if ( debug && ( stage == STAGE_PREAMBLE ) && ( preSize < PREAMBLE_THRESHOLD / x )) printf ( "Found data but preamble too short(%d/%d)\n", preSize, PREAMBLE_THRESHOLD / x );
            if (( stage == STAGE_PREAMBLE ) && ( preSize > PREAMBLE_THRESHOLD / x )) {
if ( debug ) printf ( "Start of data detected\n" );
                stage = STAGE_DATA;
            } else if ( stage != STAGE_DATA ) {
                stage = STAGE_UNKNOWN;
            }
            bad = 0;
            preSize = 0;
        } else {
            if ( bad++ > 1 ) {
                if ( stage == STAGE_DATA ) {
                    fileData [ count ].DataLength = tapeBuffer - fileData [ count ].pData;
                    fileData [ count ].period = periodSum / periodCount;
                    periodSum = 0.0;
                    periodCount  = 0;
                    count++;
printf ( "\rFound %d file%s", count, ( count > 1 ) ? "s" : "" );
                }
                stage = STAGE_UNKNOWN;
                fileData [ count ].pData = tapeBuffer;
            }
            preSize = 0;
        }

    } while ( true );

done:

    if ( stage == STAGE_DATA ) {
        fileData [ count ].DataLength = tapeBuffer - fileData [ count ].pData;
        fileData [ count ].period = periodSum / periodCount;
        count++;
printf ( "\rFound %d file%s", count, ( count > 1 ) ? "s" : "" );
    }

printf ( "\n" );

    return count;
}

int main ( int argc, char *argv [] )
{
    printf ( "TI-99/4A Cassette Utility\n" );

    int index = 1;

    if ( index >= argc ) {
        fprintf ( stderr, "No input file specified\n" );
        return -1;
    }

    char fileName [256];
    strcpy ( fileName, argv [index] );

    FILE *file = fopen ( fileName, "rb" );
    if ( file == NULL ) {
        strcat ( fileName, ".wav" );        
        file = fopen ( fileName, "rb" );
        if ( file == NULL ) {
            fprintf ( stderr, "Unable to open file '%s'\n", argv [index] );
            return -1;
        }
    }

    fseek ( file, 0, SEEK_END );
    ULONG size = ftell ( file );
    fseek ( file, 0, SEEK_SET );

    printf ( "Reading file %s\n", fileName );

    UCHAR *buffer = new UCHAR [ size ];
    fread ( buffer, 1, size, file );
    fclose ( file );

    UCHAR *pData;
    int    DataSize;

    if ( CheckHeader ( buffer, size, &pData, &DataSize, &fmt ) == false ) {
        fprintf ( stderr, "Invalid WAV file\n" );
        return -1;
    }

    if ( fmt->formatTag != 1 ) {
        fprintf ( stderr, "Unsupported data format\n" );
        return -1;
    }

    int width = fmt->nAvgBytesPerSec / fmt->nSamplesPerSec;
    switch ( fmt->nChannels ) {
        case 1 : GetSample = ( width == 1 ) ? GetSampleMono8   : GetSampleMono16;	break;
        case 2 : GetSample = ( width == 2 ) ? GetSampleStereo8 : GetSampleStereo16;	break;
    }

    printf ( "Format: %d-bit %s %d,%03dHz\n", 8 * width, ( fmt->nChannels == 1 ) ? "Mono" : "Stereo", fmt->nSamplesPerSec / 1000, fmt->nSamplesPerSec % 1000 );

    threshold = 0.15;
    cutOff    = ( width == 1 ) ? 0x01 : 0x0100;

    printf ( "Processing...\n" );
/*
    sFileData fileData [32];
    memset ( fileData, 0, sizeof ( fileData ));
    int count = LocateFiles ( pData, DataSize, fileData );

    char filename [ 256 ];
    for ( int i = 0; i < count; i++ ) {
        sprintf ( filename, "track-%02d.dat", i + 1 );
        printf ( "Reading track %2d: ", i + 1 );

        UCHAR *bitBuffer = new UCHAR [ fileData [i].DataLength ];
        int bitCount = FillBitBuffer ( fileData [i].pData, fileData [i].DataLength, fileData [i].period, bitBuffer );
        DumpBuffer ( bitBuffer, bitCount, filename );
        delete [] bitBuffer;
    }
*/

    UCHAR *bitBuffer = new UCHAR [ DataSize ];
    period = FindPeriod ( pData, DataSize );
    int bitCount = FillBitBuffer ( pData, DataSize, period, bitBuffer );
    DumpBuffer ( bitBuffer, bitCount, "tape.dat" );
    delete [] bitBuffer;

    delete [] buffer;

    return 0;
}
