//----------------------------------------------------------------------------
//
// File:        disk.cpp
// Date:        23-Feb-1998
// Programmer:  Marc Rousseau
//
// Description: A simple program to display the contents of a TI-Disk image
//              created by AnaDisk.  It also supports extracting files from
//              the disk and optionally converting them to 'standard' PC type
//              files.
//
// Copyright (c) 1994-2001 Marc Rousseau, All Rights Reserved.
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307, USA.
//
// Revision History:
//
//----------------------------------------------------------------------------

#include <ctype.h>
#include <memory.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "common.hpp"
#include "logger.hpp"
#include "diskio.hpp"
#include "fileio.hpp"
#include "option.hpp"

DBG_REGISTER ( __FILE__ );

void ConvertFilename ( char *buffer )
{
    char *end = buffer;
    while ( *end ) end++;
    while ( *--end == ' ' );
    *++end = '\0';
    if ( *buffer == '-' ) *buffer = '_';
    for ( ; *buffer; buffer++ ) {
        if ( isalnum ( *buffer )) continue;
        if ( *buffer == '_' ) continue;
        if ( *buffer == '-' ) continue;
        *buffer = '_';
    }
}

USHORT GetUSHORT ( void *_ptr )
{
    UCHAR *ptr = ( UCHAR * ) _ptr;
    return ( ptr [0] << 8 ) | ptr [1];
}

const sSector *FindSector ( cDiskMedia *disk, VIB *vib, int index )
{
    int trackSize = (( vib != NULL ) && ( vib->SectorsPerTrack != 0 )) ? vib->SectorsPerTrack : 9;

    int t = index / trackSize;
    int s = index % trackSize;
    int h = 0;

    if ( t >= MAX_TRACKS ) {
        t = 2 * MAX_TRACKS - t - 1;
        h = 1;
    }

    if ( t >= MAX_TRACKS ) {
        fprintf ( stderr, "Invalid sector index (%d)\n", index );
        exit ( - 1 );
        return NULL;
    }

    return disk->GetSector ( t, h, s );
}

void DumpSectors ( cDiskMedia *disk, VIB *vib, int sIndex, int sCount, bool isLast, sFileDescriptorRecord *FDR, bool convertFiles, FILE *outFile )
{
    bool isDisplay = (( FDR->FileStatus & INTERNAL_TYPE ) == 0 ) ? true : false;

    for ( int i = 0; i < sCount; i++ ) {

        const sSector *sector = FindSector ( disk, vib, sIndex );
        if ( sector == NULL ) {
            fprintf ( stderr, "File: %s - sector %d not found\n", FDR->FileName, sIndex );
            sIndex++;
            continue;
        }
        UCHAR *ptr = sector->Data;

        if ( convertFiles == false ) {
            fwrite ( ptr, 256, 1, outFile );
        } else {
            if ( FDR->FileStatus & PROGRAM_TYPE ) {
                int size = (( i == sCount - 1 ) && isLast ) ? FDR->EOF_Offset : 256;
                fwrite ( ptr, size, 1, outFile );
            } else if ( FDR->FileStatus & VARIABLE_TYPE ) {
                while ( *ptr != 0xFF ) {
                    if ( isDisplay == true ) {
                        int length = *ptr++;
                        fwrite ( ptr, length, 1, outFile );
                        fputs ( "\n", outFile );
                        ptr += length;
                    } else {
                        int length = *ptr + 1;
                        fwrite ( ptr, length, 1, outFile );
                        ptr += length;
                    }
                }
            } else {
                for ( int j = 0; j < FDR->RecordsPerSector; j++ ) {
                    if ( FDR->NoFixedRecords == 0 ) break;
                    FDR->NoFixedRecords--;
                    fwrite ( ptr, FDR->RecordLength, 1, outFile );
                    if ( isDisplay == true ) {
                        fputs ( "\n", outFile );
                    }
                    ptr += FDR->RecordLength;
                }
            }
        }
        sIndex++;
    }
}

void DumpFile ( cDiskMedia *disk, VIB *vib, sFileDescriptorRecord *FDR, bool convertFiles )
{
    char name [ 32 ];

    strcpy ( name, FDR->FileName );
    ConvertFilename ( name );

    // Decorate the name to match the file's type
    if ( convertFiles == true ) {
        if ( FDR->FileStatus & PROGRAM_TYPE ) {
            strcat ( name, ".PROG" );
        } else {
            sprintf ( name + strlen ( name ), ".%c%c%d",
                ( FDR->FileStatus & INTERNAL_TYPE ) ? 'I' : 'D',
                ( FDR->FileStatus & VARIABLE_TYPE ) ? 'V' : 'F',
                FDR->RecordLength );
        }
    }

    // Find a unique filename to use
    char buffer [ 80 ];
    FILE *outFile;
    for ( int i = 0; i < 255; i++ ) {
        sprintf ( buffer, i ? "%s.%03d" : "%s", name, i );
        outFile = fopen ( buffer, "rb" );
        if ( outFile == NULL ) break;
        fclose ( outFile );
    }

    const char *mode = (( convertFiles == true ) && (( FDR->FileStatus & ( INTERNAL_TYPE | PROGRAM_TYPE )) == 0 )) ? "wt" : "wb";

    // Open the file
    outFile = fopen ( buffer, mode );
    if ( outFile == NULL ) return;

    // Write the file header
    if ( convertFiles == false ) {
        char header [128];
        memset ( header, 0, 128 );
        memcpy ( header, FDR, 20 );
        fwrite ( header, 128, 1, outFile );
    }

    int size = GetUSHORT ( &FDR->TotalSectors );
    int copied = 0;
    CHAIN *chain = FDR->DataChain;
    do {
        int sector = chain->start + (( int ) ( chain->start_offset & 0x0F ) << 8 );
        int offset = (( int ) chain->offset << 4 ) + ( chain->start_offset >> 4 ) + 1;
        DumpSectors ( disk, vib, sector, offset - copied, offset == size, FDR, convertFiles, outFile );
        copied = offset;
        chain++;
    } while ( copied < size );

    fclose ( outFile );
}

void DumpFile ( cDiskMedia *disk, VIB *vib, char *name, bool convertFiles )
{
    USHORT *dirIndex = ( USHORT * ) FindSector ( disk, vib, 1 )->Data;

    for ( int i = 0; i < 128; i++ ) {

        if ( dirIndex [i] == 0 ) break;

        int index = GetUSHORT ( &dirIndex [i] );

        const sSector *sector = FindSector ( disk, vib, index );
        sFileDescriptorRecord *FDR = ( sFileDescriptorRecord * ) sector->Data;

        if ( strnicmp ( FDR->FileName, name, strlen ( name )) == 0 ) {
            DumpFile ( disk, vib, FDR, convertFiles );
            break;
        }
    }
}

void ShowDirectory ( cDiskMedia *disk, bool dumpFiles, bool convertFiles )
{
    if ( disk->GetFormat () == FORMAT_UNKNOWN ) {
        printf ( "Could not read the disk format\n" );
        return;
    }

    VIB *vib = ( VIB * ) disk->GetSector ( 0, 0, 0 )->Data;

    printf ( "Directory of %10.10s\n", vib->VolumeName );
    printf ( "\n" );
    printf ( " Filename   Sectors   Type\n" );

    USHORT *dirIndex = ( USHORT * ) FindSector ( disk, vib, 1 )->Data;

    for ( int i = 0; i < 128; i++ ) {

        if ( dirIndex [i] == 0 ) break;

        int index = GetUSHORT ( &dirIndex [i] );

        const sSector *sector = FindSector ( disk, vib, index );
        sFileDescriptorRecord *FDR = ( sFileDescriptorRecord * ) sector->Data;

        printf ( "  %10.10s", FDR->FileName );
        int size = GetUSHORT ( &FDR->TotalSectors );
        printf ( " %3d", size );
        if ( FDR->FileStatus & PROGRAM_TYPE ) {
	    printf ( "      PROGRAM    " );
        } else {
            printf ( "      %s/%s %3d", ( FDR->FileStatus & INTERNAL_TYPE ) ? "INT" : "DIS",
                                        ( FDR->FileStatus & VARIABLE_TYPE ) ? "VAR" : "FIX",
                                        FDR->RecordLength );
        }
        printf ( "\n" );

        if ( dumpFiles == true ) {
            DumpFile ( disk, vib, FDR, convertFiles );
        }
    }
}

bool FileExists ( const char *filename )
{
    FILE *file = fopen ( filename, "rb" );
    if ( file == NULL ) return false;
    fclose ( file );
    return true;
}

void PrintUsage ()
{
    FUNCTION_ENTRY ( NULL, "PrintUsage", true );

    cprintf ( "Usage: disk [options] [file]\n" );
    cprintf ( "\n" );
}

void PrintDiskInfo ( cDiskMedia &disk )
{
    int noTracks = -1;
    int noSides  = -1;
    eDiskDensity density = DENSITY_UNKNOWN;

    for ( int h = 0; h < 2; h++ ) {
        for ( int t = 0; t < MAX_TRACKS; t++ ) {
            const sTrack *track = disk.GetTrack ( t, h );
            if ( track == NULL ) continue;
            printf ( "Track: %2d  Side: %d - ", t, h );
            if ( density == DENSITY_UNKNOWN ) {
                density = track->Density;
            } else if ( density != track->Density ) {
                if ( track->Density == DENSITY_UNKNOWN ) {
                    printf ( "is not formatted\n" );
                    continue;
                } else {
                    printf ( "density changed -" );
                }
            }
            noSides  = h + 1;
            noTracks = t + 1;
            int  noSectors = 0;
            bool deletedSectors = false;
            for ( int s = 0; s < MAX_SECTORS; s++ ) {
                if ( track->Sector [s].Data == NULL ) continue;
                printf ( " %d", track->Sector [s].LogicalSector );
                noSectors++;
                if ( track->Sector [s].LogicalCylinder != t ) {
                    printf ( "  track %d sector %d has incorrect cylinder %d\n", t, s, track->Sector [s].LogicalCylinder );
                }
                if ( track->Sector [s].LogicalSide != h ) {
                    printf ( "  track %d sector %d has incorrect side %d\n", t, s, track->Sector [s].LogicalSide );
                }
                if ( track->Sector [s].Data [-1] != 0xFB ) {
                    printf ( "  track %d contains deleted sectors\n", t );
                    deletedSectors = true;
                }
                if ( track->Sector [s].Size != 1 ) {
                    printf ( "  track %d sector %d size is %d\n", t, s, 128 << track->Sector [s].Size );
                }
            }
            if (( noSectors != 9 ) && ( noSectors != 18 )) {
                printf ( " - track %d contains %d sectors", t, noSectors );
            }
            printf ( "\n" );
        }
    }

    const char *strFormat = "<Unknown>";
    switch ( disk.GetFormat ()) {
        case FORMAT_RAW_TRACK :
            strFormat = "PC99";
            break;
        case FORMAT_RAW_SECTOR :
            strFormat = "v9t9";
            break;
        case FORMAT_ANADISK :
            strFormat = "AnaDisk";
            break;
        default :
            break;
    }
    printf ( " Format: %s\n", strFormat );

    const char *txtDensity = "<Unknown>";
    switch ( density ) {
        case DENSITY_SINGLE :
            txtDensity = "Single";
            break;
        case DENSITY_DOUBLE :
            txtDensity = "Double";
            break;
        default :
            break;
    }

    printf ( " Tracks: %d\n", noTracks );
    printf ( "  Sides: %d\n", noSides );
    printf ( "Density: %s\n", txtDensity );
    printf ( "\n" );
}

int main ( int argc, char *argv[] )
{
    bool dumpFiles    = false;
    bool convertFiles = false;
    bool verboseMode  = false;
    eDiskFormat outputFormat = FORMAT_UNKNOWN;

    sOption optList [] = {
        { 'd', "dump",               OPT_VALUE_SET | OPT_SIZE_BOOL, true,              &dumpFiles,    NULL,           "Extract files to v9t9 FIAD files" },
        { 'c', "convert",            OPT_VALUE_SET | OPT_SIZE_BOOL, true,              &convertFiles, NULL,           "Convert extracted files to DOS files" },
        {  0,  "output=PC99",        OPT_VALUE_SET | OPT_SIZE_INT,  FORMAT_RAW_TRACK,  &outputFormat, NULL,           "Convert disk to PC99 format" },
        {  0,  "output=v9t9",        OPT_VALUE_SET | OPT_SIZE_INT,  FORMAT_RAW_SECTOR, &outputFormat, NULL,           "Convert disk to v9t9 DOAD format" },
        {  0,  "output=anadisk",     OPT_VALUE_SET | OPT_SIZE_INT,  FORMAT_ANADISK,    &outputFormat, NULL,           "Convert disk to AnaDisk format /w headers" },
        { 'v', "verbose",            OPT_VALUE_SET | OPT_SIZE_BOOL, true,              &verboseMode,  NULL,           "Display information about the disk image" }
    };

    if ( argc == 1 ) {
        PrintHelp ( SIZE ( optList ), optList );
        return 0;
    }

    printf ( "TI-99/4A Diskette Viewer\n" );

    int index = 1;
    index = ParseArgs ( index, argc, argv, SIZE ( optList ), optList );

    if ( index >= argc ) {
        fprintf ( stderr, "No input file specified\n" );
        return -1;
    }

    const char *validName = LocateFile ( "disks", argv [index++] );
    if ( validName == NULL ) {
        fprintf ( stderr, "Unable to open file '%s'\n", argv [index-1] );
        return -1;
    }

    cDiskMedia disk ( validName );

    if ( disk.GetFormat () == FORMAT_UNKNOWN ) {
        fprintf ( stderr, "File '%s' contains an unrecognized disk format\n", disk.GetName ());
        return -1;
    }

    if ( verboseMode == true ) {
        PrintDiskInfo ( disk );
        const char *strFormat = "<Unknown>";
        switch ( disk.GetFormat ()) {
            case FORMAT_RAW_TRACK :
                strFormat = "PC99";
                break;
            case FORMAT_RAW_SECTOR :
                strFormat = "PC99";
                break;
            case FORMAT_ANADISK :
                strFormat = "AnaDisk";
                break;
            default :
                break;
        }
        printf ( "Format: %s\n", strFormat );
    }

    ShowDirectory ( &disk, dumpFiles, convertFiles );

    VIB *vib = ( VIB * ) disk.GetSector ( 0, 0, 0 )->Data;

    while ( index < argc ) {
        DumpFile ( &disk, vib, argv[index++], convertFiles );
    }

    if ( outputFormat != FORMAT_UNKNOWN ) {
        if ( disk.SaveFile ( outputFormat, true ) == false ) {
            fprintf ( stderr, "Unable to write to file '%s'\n", disk.GetName ());
            return -1;
        }
    }

    return 0;
}
